/*****************************************************************************

        DiffEvol.cpp
        Copyright (c) 2001-2003 Laurent de Soras

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130) // "'operator' : logical operation on address of string constant"
	#pragma warning (1 : 4223) // "nonstandard extension used : non-lvalue array converted to pointer"
	#pragma warning (1 : 4705) // "statement has no effect"
	#pragma warning (1 : 4706) // "assignment within conditional expression"
	#pragma warning (4 : 4786) // "identifier was truncated to '255' characters in the debug information"
	#pragma warning (4 : 4800) // "forcing value to bool 'true' or 'false' (performance warning)"
	#pragma warning (4 : 4355) // "'this' : used in base member initializer list"
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"DiffEvol.h"
#include	"DiffEvolFtorInterface.h"
#include	"DiffEvolPopulation.h"

#include	<cassert>



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: ctor
Description:
	Build and initialize the DE object. The min/max array must respect the
	following requirements:
	- Both arrays should have the same length, > 0
	- For every i, min [i] <= max [i]
	- The length of these arrays sets the number of parameters per population.
Input parameters:
	- nbr_pop: Number of populations. Must be >= 5.
	- min_arr: Array of the minimum values for the initial generation
	- max_arr: Array of the maximum values for the initial generation
Throws: exceptions related to memory allocation for vectors
==============================================================================
*/

DiffEvol::DiffEvol (int nbr_pop, const DiffEvolPopulation &min_arr, const DiffEvolPopulation &max_arr)
:	_nbr_pop (nbr_pop)
,	_nbr_var (min_arr.size ())
,	_best_pop (-1)
{
	assert (_nbr_pop >= 5);
	assert (_nbr_var > 0);
	assert (max_arr.size () == _nbr_var);

	init (&min_arr [0], &max_arr [0], 0);
}



/*
==============================================================================
Name: ctor
Description:
	Build and initialize the DE object. The min/max array must respect the
	following requirements:
	- Both arrays should have the same length, > 0
	- For every i, min [i] <= max [i]
Input parameters:
	- nbr_pop: Number of populations. Must be >= 5.
	- nbr_var: Number of variables, size of the population and arrays. > 0
	- min_arr: Array of the minimum values for the initial generation
	- max_arr: Array of the maximum values for the initial generation
Throws: exceptions related to memory allocation for vectors
==============================================================================
*/

DiffEvol::DiffEvol (int nbr_pop, int nbr_var, const double min_arr [], const double max_arr [])
:	_nbr_pop (nbr_pop)
,	_nbr_var (nbr_var)
,	_best_pop (-1)
{
	assert (_nbr_pop >= 5);
	assert (_nbr_var > 0);
	assert (min_arr != 0);
	assert (max_arr != 0);

	init (&min_arr [0], &max_arr [0], 0);
}



/*
==============================================================================
Name: ctor
Description:
	Same things than other ctors, gives a clue to the DE for the initial best
	population, so the algorithm does not start from total randomness.
Input parameters:
	- nbr_pop: Number of populations. Must be >= 5.
	- min_arr: Array of the minimum values for the initial generation
	- max_arr: Array of the maximum values for the initial generation
	- best_arr: Set of initial values.
Throws: exceptions related to memory allocation for vectors
==============================================================================
*/

DiffEvol::DiffEvol (int nbr_pop, const DiffEvolPopulation &min_arr, const DiffEvolPopulation &max_arr, const DiffEvolPopulation &best_arr)
:	_nbr_pop (nbr_pop)
,	_nbr_var (min_arr.size ())
,	_best_pop (-1)
{
	assert (_nbr_pop >= 5);
	assert (_nbr_var > 0);
	assert (max_arr.size () == _nbr_var);
	assert (best_arr.size () == _nbr_var);

	init (&min_arr [0], &max_arr [0], &best_arr [0]);
}



/*
==============================================================================
Name: ctor
Description:
	Guess what...
Throws: exceptions related to memory allocation for vectors
==============================================================================
*/

DiffEvol::DiffEvol (int nbr_pop, int nbr_var, const double min_arr [], const double max_arr [], const double best_arr [])
:	_nbr_pop (nbr_pop)
,	_nbr_var (nbr_var)
,	_best_pop (-1)
{
	assert (_nbr_pop >= 5);
	assert (_nbr_var > 0);
	assert (min_arr != 0);
	assert (max_arr != 0);
	assert (best_arr != 0);

	init (&min_arr [0], &max_arr [0], &best_arr [0]);
}



/*
==============================================================================
Name: evolve
Description:
	Compute the next generation, trying to find a population which cost is
	lower than the previous one. Note: coefficient for the original population
	is 1.0 minus sum of gain_*. The coefficients can be negative as well as
	positive, and be higher than 1.
Input parameters:
	- gain_best: Coefficient applyed to the previous best population
	- gain_r1: Coefficient for 1st random population
	- gain_r2: Coefficient for 2nd random population
	- gain_r3: Coefficient for 3rd random population
	- cr: Crossing-over amount, in [0 ; 1]
Input/output parameters:
	- func: Object implementing the cost function. It shall accept populations
		of the size given at the DiffEvol creation.
Returns: The new best cost.
Throws: Nothing
==============================================================================
*/

double	DiffEvol::evolve (DiffEvolFtorInterface &func, double gain_best, double gain_r1, double gain_r2, double gain_r3, double cr)
{
	assert (cr >= 0);
	assert (cr <= 1);

	if (_best_pop < 0)
	{
		find_best (func);
	}

	const double	gain_r0 = 1 - gain_best - gain_r1 - gain_r2 - gain_r3;
	int				new_best_pop = _best_pop;
	double			new_best_cost = _cur_gen [_best_pop]._cost;

	for (int pop = 0; pop < _nbr_pop; ++pop)
	{
		// Find 3 different populations randomly
		int				r1;
		do
		{
			r1 = static_cast <int> (_rnd.gen_uint32 () % _nbr_pop);
		}
		while (   r1 == pop
		       || r1 == _best_pop);

		int				r2;
		do
		{
			r2 = static_cast <int> (_rnd.gen_uint32 () % _nbr_pop);
		}
		while (   r2 == pop
		       || r2 == _best_pop
		       || r2 == r1);

		int				r3;
		do
		{
			r3 = static_cast <int> (_rnd.gen_uint32 () % _nbr_pop);
		}
		while (   r3 == pop
		       || r3 == _best_pop
		       || r3 == r1
		       || r3 == r2);

		// Generate trial vector with crossing-over
		int				var = static_cast <int> (_rnd.gen_uint32 () % _nbr_var);
		int				var_cnt = 0;
		double			fr;
		do
		{
			_next_gen [pop]._pop [var] =
				  _cur_gen [      pop]._pop [var] * gain_r0
				+ _cur_gen [       r1]._pop [var] * gain_r1
				+ _cur_gen [       r2]._pop [var] * gain_r2
				+ _cur_gen [       r3]._pop [var] * gain_r3
				+ _cur_gen [_best_pop]._pop [var] * gain_best;

			var = (var + 1) % _nbr_var;
			++var_cnt;
			fr = _rnd.gen_double ();
		}
		while (var_cnt < _nbr_var && fr < cr);
		while (var_cnt < _nbr_var)
		{
			_next_gen [pop]._pop [var] = _cur_gen [pop]._pop [var];
			var = (var + 1) % _nbr_var;
			++var_cnt;
		}

		// Evaluate the new population
		_next_gen [pop]._cost = func (_next_gen [pop]._pop);
		_next_gen [pop]._valid_cost_flag = true;

		// Not better, discard it and keep the old one.
		if (_next_gen [pop]._cost >= _cur_gen [pop]._cost)
		{
			_next_gen [pop] = _cur_gen [pop];
		}

		// Better, keep it.
		else
		{
			// New best ?
			if (_next_gen [pop]._cost < new_best_cost)
			{
				new_best_pop = pop;
				new_best_cost = _next_gen [pop]._cost;
			}
		}
	}

	_best_pop = new_best_pop;
	_cur_gen = _next_gen;

	return (new_best_cost);
}



/*
==============================================================================
Name: get_best_population
Description:
	Return the current best population. This function must not be called
	before the first call to evolve ().
Returns: A reference on the population. It remains valid until subsequent
	call to evolve() function.
Throws: Nothing
==============================================================================
*/

const DiffEvolPopulation &	DiffEvol::get_best_population () const
{
	assert (_best_pop >= 0);

	return (_cur_gen [_best_pop]._pop);
}



/*
==============================================================================
Name: get_best_cost
Description:
	Return the cost evaluation for the current best population. This function
	must not be called before the first call to evolve ().
Returns: The cost.
Throws: Nothing
==============================================================================
*/

double	DiffEvol::get_best_cost () const
{
	assert (_best_pop >= 0);
	assert (_cur_gen [_best_pop]._valid_cost_flag);

	return (_cur_gen [_best_pop]._cost);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



DiffEvol::EvaluatedPopulation::EvaluatedPopulation ()
:	_valid_cost_flag (false)
{
}



void	DiffEvol::init (const double min_arr [], const double max_arr [], const double best_arr [])
{
	// Size all the arrays
	_cur_gen.resize (_nbr_pop);
	_next_gen.resize (_nbr_pop);
	for (int pop = 0; pop < _nbr_pop; ++pop)
	{
		_cur_gen [pop]._pop.resize (_nbr_var);
		_next_gen [pop]._pop.resize (_nbr_var);
	}

	// Initialize populations with random values
	randomize_population (min_arr, max_arr);

	// Introduce the "best" population if it is provided
	if (best_arr != 0)
	{
		for (int var = 0; var < _nbr_var; ++var)
		{
			_cur_gen [0]._pop [var] = best_arr [var];
		}
		_best_pop = 0;
	}
}



void	DiffEvol::randomize_population (const double min_arr [], const double max_arr [])
{
	for (int var = 0; var < _nbr_var; ++var)
	{
		assert (min_arr [var] <= max_arr [var]);

		const double	offset = min_arr [var];
		const double	mul = max_arr [var] - min_arr [var];

		for (int pop = 0; pop < _nbr_pop; ++pop)
		{
			const double	val_rnd = _rnd.gen_double ();
			_cur_gen [pop]._pop [var] = val_rnd * mul + offset;
		}
	}

	_best_pop = -1;
}



double	DiffEvol::find_best (DiffEvolFtorInterface &func)
{
	if (_best_pop < 0)
	{
		_best_pop = 0;
	}

	if (! _cur_gen [_best_pop]._valid_cost_flag)
	{
		_cur_gen [_best_pop]._cost = func (_cur_gen [_best_pop]._pop);
		_cur_gen [_best_pop]._valid_cost_flag = true;
	}

	double			best_cost = _cur_gen [_best_pop]._cost;
	for (int pop = 0; pop < _nbr_pop; ++pop)
	{
		if (pop != _best_pop)
		{
			const double	cur_cost = func (_cur_gen [pop]._pop);

			_cur_gen [pop]._cost = cur_cost;
			_cur_gen [pop]._valid_cost_flag = true;

			if (cur_cost < best_cost)
			{
				_best_pop = pop;
				best_cost = cur_cost;
			}
		}
	}

	return (best_cost);
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
