/*****************************************************************************

        Differential Evolution
        Copyright (c) 2001-2003 Laurent de Soras

Test for DiffEvol class. See comments in DiffEvol.h

TestFunction implements a "secret" function we'll try to approximate with
a 4th-order polynomial. For the need of the test, this secret function is
also a 4th-order polynomial so we can check the results easily.

--- Legal stuff ---

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*Tab=3***********************************************************************/



#include	"DiffEvol.h"
#include	"DiffEvolFtorInterface.h"
#include	"DiffEvolPopulation.h"

#include	<cassert>
#include	<cmath>
#include	<cstdio>



class TestFunction
:	public DiffEvolFtorInterface
{
public:

	enum { ORDER = 4 };

	virtual double	operator () (const DiffEvolPopulation &pop);

	static const double
					_real_coef [ORDER + 1];
};

const double	TestFunction::_real_coef [ORDER + 1] =
{
	2.25, -4.9, 3.58, 0.7, -0.169
};

double	TestFunction::operator () (const DiffEvolPopulation &pop)
{
	assert (pop.size () == ORDER + 1);

	double			err_sum = 0;

	// Integration of (ftest (x) - fref (x)) ^ 2 over [-5; 5]
	const double	step = 1.0 / 128;
	for (double x = -5; x <= 5; x += step)
	{
		double			y_pop = 0;
		double			y_ref = 0;

		for (int i = ORDER; i >= 0; --i)
		{
			y_pop = y_pop * x + pop [i];
			y_ref = y_ref * x + _real_coef [i];
		}

		const double	err_val = y_pop - y_ref;
		err_sum += err_val * err_val;
	}

	using namespace std;

	// Log has no impact on performances here but it's easier to follow...
	err_sum += 1e-200;	// To ensure it's strictly positive
	const double	cost = log (err_sum);

	return (cost);
}



int	main ()
{
	using namespace std;

	// Same range for every coef.
	DiffEvolPopulation	min_arr (TestFunction::ORDER + 1, -100);
	DiffEvolPopulation	max_arr (TestFunction::ORDER + 1, +100);

	// 100 populations per generation
	DiffEvol			de (100, min_arr, max_arr);
	TestFunction	func;

	// Here, the exact coefficients are found after about 700 iterations
	for (long pass = 0; pass < 300; ++pass)
	{
		de.evolve (func, 0, -0.7, 0.7, 1.0, 1.0);
		const double	cost = de.get_best_cost ();
		printf ("Pass%8ld : %#+5g\n", pass, cost);
	}

	// Print result
	printf ("Theoric / Found / Error\n");
	const DiffEvolPopulation &	best_pop = de.get_best_population ();
	for (int i = 0; i <= TestFunction::ORDER; ++i)
	{
		const double	error =
			fabs (TestFunction::_real_coef [i] - best_pop [i]);
		printf (
			"%+3.9f %+3.9f %#5g\n",
			TestFunction::_real_coef [i],
			best_pop [i],
			error
		); 
	}

	return (0);
}